import 'dart:async';
import 'dart:convert';
import 'package:http/http.dart';
import 'package:senior_authentication/src/models/user.dart';
import 'package:senior_authentication/src/utils/utils.dart';

import './models/models.dart';
import './utils/exceptions.dart';
import './utils/preference_storage.dart';

/// Indicates the possible authentication status of senior X Platform.
enum AuthenticationStatus {
  /// Initial status
  unknown,

  /// Status when user is authenticated on senior X Platform.
  authenticated,

  /// Status when user is not authenticated on senior X Platform.
  unauthenticated,
}

///
class AuthenticationRepository {
  /// {@macro authentication_repository}
  ///
  /// If no [urlBase] is provided, production enpoint will be used.
  factory AuthenticationRepository(
      {String urlBase =
          'https://platform.senior.com.br/t/senior.com.br/bridge/1.0/rest'}) {
    _instance ??= AuthenticationRepository._(urlBase);
    return _instance!;
  }

  AuthenticationRepository._(
    this.urlBase,
  );

  /// senior X Platform communication URL.
  final String urlBase;

  static AuthenticationRepository? _instance;

  final Map<String, String> _headers = {'Content-type': 'application/json'};
  final _controller = StreamController<AuthenticationStatus>();

  /// Gets the actual authentication status.
  ///
  /// View [AuthenticationStatus] for details.
  Stream<AuthenticationStatus> get status async* {
    yield AuthenticationStatus.unauthenticated;
    yield* _controller.stream;
  }

  /// Returns the current cached token.
  /// Defaults to [Token.empty] if there is no cached token.
  Future<Token> get currentToken async {
    return await PreferenceStorage.getToken();
  }

  //// Returns the current logged user
  Future<User?> get loggedUser async {
    return await PreferenceStorage.getLoggedUser();
  }

  //// Get info from user requestings /getUser on platform
  Future<User?> getUser(String accessToken,
      {bool autoRefreshToken = false}) async {
    try {
      final headers = {
        ..._headers,
        'Authorization': AuthenticationUtils.parseTokenValue(accessToken)
      };
      final response = await get(
          Uri.parse('$urlBase/platform/user/queries/getUser'),
          headers: headers);

      if (response.statusCode == 403) {
        PreferenceStorage.clearLoggedUser();
        throw GetUserInvalidCredentials();
      } else if (response.statusCode == 200) {
        var user = User.fromJson(response.body);
        PreferenceStorage.saveLoggedUser(user);
        return user;
      }
    } on GetUserInvalidCredentials {
      if (autoRefreshToken) {
        // TODO: auto refresh token when 403

      }
    } catch (_) {
      throw GetUserFailure();
    }
  }

  /// Gets the login settings of the tenant associated with the tenantDomain
  ///
  /// Throws a [TenantLogInSettingsFailure] if an exception occurs.
  Future<TenantLoginSettings> getTenantLogInSettings(
      String tenantDomain, String? scope) async {
    try {
      final response = await post(
        Uri.parse(
            '$urlBase/platform/authentication/queries/tenantLoginSettings'),
        headers: _headers,
        body: json.encode({'tenantDomain': tenantDomain, 'scope': scope ?? ''}),
      );

      if (response.statusCode == 400 ||
          response.statusCode == 401 ||
          response.statusCode == 404) throw TenantLogInSettingsIncorrect();

      if (response.statusCode == 200)
        return TenantLoginSettings.fromJson(response.body);

      throw TenantLogInSettingsFailure();
    } catch (_) {
      rethrow;
    }
  }

  /// Signs in with the provided [UserLogin] infos.
  ///
  /// Throws a [LogInWithEmailAndPasswordFailure] if an exception occurs.
  /// Throws a [LogInWithEmailAndPasswordIncorrect] if user or password
  /// are incorrect.
  Future<void> logInWithEmailAndPassword(UserLogin userLogin) async {
    try {
      var response = await post(
        Uri.parse('$urlBase/platform/authentication/actions/login'),
        headers: _headers,
        body: userLogin.toJson(),
      );

      if (response.statusCode == 401) {
        throw LogInWithEmailAndPasswordIncorrect();
      }

      // ignore: avoid_dynamic_calls
      var token = Token.fromJson(json.decode(response.body)['jsonToken']);

      PreferenceStorage.saveToken(token);
      _controller.add(AuthenticationStatus.authenticated);
    } on LogInWithEmailAndPasswordIncorrect {
      rethrow;
    } catch (error) {
      throw LogInWithEmailAndPasswordFailure();
    }
  }

  /// Starts the Sign In with SAML Flow.
  ///
  /// Throws a [LogInSamlFailure] if an exception occurs.
  void logInSaml(String tokenStr) {
    try {
      var token = Token.fromJson(tokenStr);
      PreferenceStorage.saveToken(token);
      _controller.add(AuthenticationStatus.authenticated);
    } catch (error) {
      throw LogInSamlFailure();
    }
  }

  /// Calls the LogOut
  ///
  /// Throws a [LogOutFailure] if an exception occurs.
  /// Throws a [LogOutIncorrect] if token is incorrect.
  Future<void> logOut() async {
    var token = await currentToken;

    if (token.isNotEmpty) {
      try {
        var accessToken = token.access_token;
        var response = await post(
          Uri.parse('$urlBase/platform/authentication/actions/logout'),
          headers: {
            'Content-type': 'application/json',
            'Authorization': 'Bearer $accessToken'
          },
          body: json.encode({'accessToken': accessToken}),
        );

        if (response.statusCode == 401) throw LogOutIncorrect();

        PreferenceStorage.clearToken();
        _controller.add(AuthenticationStatus.unauthenticated);
      } catch (error) {
        throw LogOutFailure();
      }
    } else {
      _controller.add(AuthenticationStatus.unauthenticated);
    }
  }

  /// Dispose method is exposed so that the authentication status
  /// StreamController can be closed when it is no longer needed.
  void dispose() => _controller.close();
}
